#define TGVOIP_USE_CALLBACK_AUDIO_IO
#include "VoIP.h"
#include <iostream>
#include <VoIPController.h>
#include <VoIPServerConfig.h>
#include <logging.h>
#include <fstream>
#include <logging.h>

using namespace VoIPNamespace;

void VoIP::initVoIPController()
{
	inst = new VoIPController();

	outputFile = NULL;

	inst->implData = (void*)this;
	VoIPController::Callbacks callbacks;
	callbacks.connectionStateChanged = [](VoIPController* controller, int state) {
		((VoIP*)controller->implData)->state = state;
		if (state == STATE_FAILED)
		{
			std::cout << "STATE_FAILED" << std::endl;
			((VoIP*)controller->implData)->deinitVoIPController();
		}
		if (state == STATE_ESTABLISHED)
		{
			std::cout << "STATE_ESTABLISHED" << std::endl;	
		}
		if (state == STATE_WAIT_INIT_ACK)
		{
			std::cout << "STATE_WAIT_INIT_ACK" << std::endl;
		}
		if (state == STATE_WAIT_INIT)
		{
			std::cout << "STATE_WAIT_INIT" << std::endl;
		}
	};
	callbacks.signalBarCountChanged = NULL;
	callbacks.groupCallKeySent = NULL;
	callbacks.groupCallKeyReceived = NULL;
	callbacks.upgradeToGroupCallRequested = NULL;
	inst->SetCallbacks(callbacks);
	inst->SetAudioDataCallbacks(
		[this](int16_t* buffer, size_t size) { this->sendAudioFrame(buffer, size); },
		[this](int16_t* buffer, size_t size) { this->recvAudioFrame(buffer, size); });
}


void VoIP::sendAudioFrame(int16_t* data, size_t size)
{
	MutexGuard m(inputMutex);
	//this->wavReader->ReadSamples(size, data);
	if (this->inputFile != NULL)
	{
		size_t r = fread(data, sizeof(int16_t), size, this->inputFile);

		if (r != size)
		{
			std::cerr << "COULD NOT READ DATA FROM FILE" << endl;
			//LOGE("COULD NOT WRITE DATA TO FILE");
		}
	}
	//std::cout << "sendAudioFrame" << endl;
	/*if (!this->inputFiles.empty())
	{
		if ((readInput = fread(data, sizeof(int16_t), size, this->inputFiles.front())) != size)
		{
			fclose(this->inputFiles.front());
			this->inputFiles.pop();
			memset(data + (readInput % size), 0, size - (readInput % size));
		}
		this->playing = true;
	}
	else
	{
		this->playing = false;
		if (!this->holdFiles.empty())
		{
			if ((readInput = fread(data, sizeof(int16_t), size, this->holdFiles.front())) != size)
			{
				fseek(this->holdFiles.front(), 0, SEEK_SET);
				this->holdFiles.push(this->holdFiles.front());
				this->holdFiles.pop();
				memset(data + (readInput % size), 0, size - (readInput % size));
			}
		}
	}*/
}

void VoIP::recvAudioFrame(int16_t* data, size_t size)
{
	//std::cout << "recvAudioFrame" << endl;
	MutexGuard m(outputMutex);
	//this->wavWriter->WriteSamples(data, size);
	if (this->outputFile != NULL)
	{
		if (fwrite(data, sizeof(int16_t), size, this->outputFile) != size)
		{
			//LOGE("COULD NOT WRITE DATA TO FILE");
			std::cerr << "COULD NOT WRITE DATA TO FILE" << endl;
		}
	}
}


void VoIP::deinitVoIPController()
{
	if (callState != CALL_STATE_ENDED)
	{
		callState = CALL_STATE_ENDED;
		if (inst)
		{
			inst->Stop();
			delete inst;
			inst = NULL;
		}

		while (holdFiles.size())
		{
			fclose(holdFiles.front());
			holdFiles.pop();
		}
		while (inputFiles.size())
		{
			fclose(inputFiles.front());
			inputFiles.pop();
		}
		unsetOutputFile();
		unsetInputFile();
	}
}

std::string VoIPNamespace::VoIP::GetDebugString()
{
	return inst->GetDebugString();
}

std::string VoIPNamespace::VoIP::GetDebugLog()
{
	return inst->GetDebugLog();
}

void VoIP::unsetOutputFile()
{
	if (outputFile == NULL)
	{
		return;
	}

	MutexGuard m(outputMutex);
	fflush(outputFile);
	fclose(outputFile);
	outputFile = NULL;

	return;
}

void VoIP::unsetInputFile()
{
	if (inputFile == NULL)
	{
		return;
	}

	MutexGuard m(inputMutex);
	fflush(inputFile);
	fclose(inputFile);
	inputFile = NULL;

	return;
}

void VoIP::setInputFile(const char* filename)
{
	MutexGuard m(inputMutex);

	if (inputFile != NULL)
	{
		fclose(inputFile);
		inputFile = NULL;
	}
	inputFile = fopen(filename, "rb");
	if (inputFile == NULL)
	{
		std::cerr << "Could not open file!" << endl;
		return;
	}


	/*FILE* tmp = fopen(filename, "rb");

	if (tmp == NULL)
	{
		std::cerr << "Could not open file!" << endl;
		return;
	}

	MutexGuard m(inputMutex);
	inputFiles.push(tmp);*/

}

void VoIP::setConfigFile(const char* filename, const char* logfile)
{
	std::string cfg_root;
	std::ifstream config_file(filename);

	if (!config_file)
		cout << "\n input file not opened successfully ";

	config_file >> cfg_root;

	//cout << cfg_root;

	std::string jsonError;
	json11::Json cmdjson = json11::Json::parse(cfg_root, jsonError);

	if (!jsonError.empty()) {

		fprintf(stderr, "initConfig ERR: can't parse json\n");
		return;
	}

	VoIPController::Config cfg;
	cfg.recvTimeout = 30.0;
	cfg.initTimeout = 30.0;
	cfg.dataSaving = DATA_SAVING_NEVER;
	cfg.enableAEC = false;
	cfg.enableNS = false;
	cfg.enableAGC = false;
	cfg.enableCallUpgrade = false;

	cfg.logFilePath = std::string(logfile);
	//cfg.statsDumpFilePath = statsDumpFilePath;

	ServerConfig::GetSharedInstance()->Update(cfg_root);

	
	inst->SetConfig(cfg);

	

	/*char* key = (char*)malloc(256);
	memcpy(key, self["configuration"]["auth_key"], 256);
	inst->SetEncryptionKey(key, (bool)self["internalStorage"]["creator"]);
	free(key);

	vector<Endpoint> eps;
	Php::Value endpoints = self["configuration"]["endpoints"];
	for (int i = 0; i < endpoints.size(); i++)
	{
		string ip = endpoints[i]["ip"];
		string ipv6 = endpoints[i]["ipv6"];
		string peer_tag = endpoints[i]["peer_tag"];

		IPv4Address v4addr(ip);
		IPv6Address v6addr("::0");
		unsigned char* pTag = (unsigned char*)malloc(16);

		if (ipv6 != "")
		{
			v6addr = IPv6Address(ipv6);
		}

		if (peer_tag != "")
		{
			memcpy(pTag, peer_tag.c_str(), 16);
		}
		eps.push_back(Endpoint(endpoints[i]["id"], (int32_t)endpoints[i]["port"], v4addr, v6addr, Endpoint::Type::UDP_RELAY, pTag));
		free(pTag);
	}

	inst->SetRemoteEndpoints(eps, (bool)self["internalStorage"]["protocol"]["udp_p2p"], (int)self["internalStorage"]["protocol"]["max_layer"]);
	inst->SetNetworkType(self["configuration"]["network_type"]);
	if (self["configuration"]["proxy"])
	{
		inst->SetProxy(self["configuration"]["proxy"]["protocol"], self["configuration"]["proxy"]["address"], (int32_t)self["configuration"]["proxy"]["port"], self["configuration"]["proxy"]["username"], self["configuration"]["proxy"]["password"]);
	}*/
}

void VoIPNamespace::VoIP::setLogFile(const wchar_t* logfile)
{
}

void VoIP::SetRemoteEndpoints(vector<Endpoint> endpoints, bool allow_p2p, int max_layer)
{
	inst->SetRemoteEndpoints(endpoints, allow_p2p, max_layer);
}

void VoIP::Start()
{
	inst->Start();
}

void VoIP::Connect()
{
	inst->Connect();
}


void VoIP::setOutputFile(const char* filename)
{
	

	MutexGuard m(outputMutex);

	if (outputFile != NULL)
	{
		fclose(outputFile);
		outputFile = NULL;
	}
	outputFile = fopen(filename, "wb");
	if (outputFile == NULL)
	{
		std::cerr << "Could not open file!" << endl;
		return;
	}
	
}

void VoIP::SetEncryptionKey(char* key, bool isOutgoing)
{
	inst->SetEncryptionKey(key, isOutgoing);
	
}

//void VoIP::initConfig()
//{
//
//	
//}
