
//#include "../StaticLib1/webrtc_dsp/common_audio/wav_file.h"

#include "VoIP.h"
#include <iostream>
#include <VoIPController.h>
#include <VoIPServerConfig.h>
#include <logging.h>
#include <fstream>
#include <climits>
#include <string>
#include <algorithm>
#include <chrono>
#include <thread>


using namespace VoIPNamespace;

static const unsigned char base64_table[65] =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";


class InputParser {
public:
	InputParser(int& argc, char** argv) {
		for (int i = 1; i < argc; ++i)
			this->tokens.push_back(std::string(argv[i]));
	}
	/// @author iain
	const std::string& getCmdOption(const std::string& option) const {
		std::vector<std::string>::const_iterator itr;
		itr = std::find(this->tokens.begin(), this->tokens.end(), option);
		if (itr != this->tokens.end() && ++itr != this->tokens.end()) {
			return *itr;
		}
		static const std::string empty_string("");
		return empty_string;
	}
	/// @author iain
	bool cmdOptionExists(const std::string& option) const {
		return std::find(this->tokens.begin(), this->tokens.end(), option)
			!= this->tokens.end();
	}
private:
	std::vector <std::string> tokens;
};



static int voip_b64_decode(std::string src, u_char* target, int targsize)
{
	unsigned char dtable[256], * pos, block[4];
	int pad = 0;

	memset(dtable, 0x80, 256);
	for (size_t i = 0; i < sizeof(base64_table) - 1; i++)
		dtable[base64_table[i]] = (unsigned char)i;
	dtable['='] = 0;

	int count = 0;
	for (char& c : src) {
		if (dtable[(unsigned int)c] != 0x80)
			count++;
	}
	if (count == 0 || count % 4)
		return -1;

	pos = target;
	count = 0;
	for (char& c : src) {
		unsigned char tmp = dtable[(unsigned int)c];
		if (tmp == 0x80)
			continue;

		if (c == '=')
			pad++;
		block[count] = tmp;
		count++;
		if (count == 4) {
			*pos++ = (block[0] << 2) | (block[1] >> 4);
			*pos++ = (block[1] << 4) | (block[2] >> 2);
			*pos++ = (block[2] << 6) | block[3];
			count = 0;
			if (pad) {
				if (pad == 1)
					pos--;
				else if (pad == 2)
					pos -= 2;
				else {
					/* Invalid padding */
					return -3;
				}
				break;
			}
		}

		if ((pos - target) > targsize)
			return -2;
	}

	return 0;
}



template <typename T>
T swap_endian(T u)
{
	static_assert (CHAR_BIT == 8, "CHAR_BIT != 8");

	union
	{
		T u;
		unsigned char u8[sizeof(T)];
	} source, dest;

	source.u = u;

	for (size_t k = 0; k < sizeof(T); k++)
		dest.u8[k] = source.u8[sizeof(T) - k - 1];

	return dest.u;
}

std::vector<char> HexToBytes(const std::string& hex) {
	std::vector<char> bytes;

	for (unsigned int i = 0; i < hex.length(); i += 2) {
		std::string byteString = hex.substr(i, 2);
		char byte = (char)strtol(byteString.c_str(), NULL, 16);
		bytes.push_back(byte);
	}

	return bytes;
}

int char2int(char input)
{
	if (input >= '0' && input <= '9')
		return input - '0';
	if (input >= 'A' && input <= 'F')
		return input - 'A' + 10;
	if (input >= 'a' && input <= 'f')
		return input - 'a' + 10;
	throw std::invalid_argument("Invalid input string");
}

// This function assumes src to be a zero terminated sanitized string with
// an even number of [0-9a-f] characters, and target to be sufficiently large
void hex2bin(const char* src, char* target)
{
	while (*src && src[1])
	{
		*(target++) = char2int(*src) * 16 + char2int(src[1]);
		src += 2;
	}
}

int main(int argc, char* argv[])
{
	
	std::cout << argv[0] << std::endl;
	if (argc < 13) {
		std::cerr << "Usage: " << argv[0] << " reflector:port tag_caller_hex -k encryption_key_hex -i /path/to/sound_A.ogg -o /path/to/sound_output_B.ogg -c config.json -r caller" << std::endl;
		return 1;
	}
	
	VoIP voip;
	const char* caller = nullptr;
	const char* reader_filename = nullptr;
	const char* writer_filename = nullptr;
	InputParser input(argc, argv);

	//-r caller
	if (input.cmdOptionExists("-r")) {
		caller = input.getCmdOption("-r").c_str();
		std::cout << "-r: " << caller << std::endl;
	}

	if (strcmp(caller, "caller") == 0)
	{
		reader_filename = "sound_A.wav";
		writer_filename = "sound_output_B.wav";
	}

	if (strcmp(caller, "callee") == 0)
	{
		reader_filename = "sound_B.wav";
		writer_filename = "sound_output_A.wav";
	}

	
	//voip.wavReader = new webrtc::WavReader(reader_filename);

	//webrtc::WavReader wavReader2("sound_B.wav");
	//voip.wavWriter = new webrtc::WavWriter(writer_filename, 48000, 1);
	//webrtc::WavWriter wavWriter2("sound_output_B.wav", 48000, 1);
	
	

	voip.initVoIPController();
	//voip2.initVoIPController();
	
	

	//reflector:port
	string addr = string(argv[1]);
	size_t found = addr.find_first_of(":");
	string host_str = addr.substr(0, found);
	string port_str = addr.substr(found + 1, addr.length() - found);
	std::cout << "host: " << host_str << std::endl;
	std::cout << "port: " << port_str << std::endl;
	//const IPv4Address reflector_ip = IPv4Address(inet_addr(host_str.c_str()));
	const IPv4Address reflector_ip = IPv4Address(host_str.c_str());
	std::cout << "reflector_ip: " << reflector_ip.addr << std::endl;
	//if (reflector_ip.IsEmpty())
	//	std::cout << "reflector ip IsEmpty" << std::endl;

	
	//tag_caller_hex
	std::string tag_caller_hex = std::string(argv[2]);
	std::cout << "tag_caller_hex: " << tag_caller_hex << std::endl;

	

	//-k encryption_key_hex
	if (input.cmdOptionExists("-k")) {
		// Do stuff
		const std::string& encryption_key_hex = input.getCmdOption("-k");
		std::cout << "encryption_key_hex: " << encryption_key_hex << std::endl;
		

		std::vector<char> key2 = HexToBytes(encryption_key_hex);
		char* key = reinterpret_cast<char*> (&key2[0]);


		
		if (strcmp(caller, "caller") == 0) 
		{
			//voip.SetEncryptionKey((char*)encryption_key_hex.c_str(), true);
			voip.SetEncryptionKey(key, true);
			//voip.SetEncryptionKey(key, true);
			std::cout << "SetEncryptionKey1" << std::endl;
		}
		
		if(strcmp(caller, "callee") == 0)
		{
			//voip.SetEncryptionKey((char*)encryption_key_hex.c_str(), false);
			voip.SetEncryptionKey(key, false);
			//voip.SetEncryptionKey(key, false);
			std::cout << "SetEncryptionKey2" << std::endl;
		}
		
		//voip2.SetEncryptionKey(key, false);
		//free(key);
		//std::cout << "key: " << key << std::endl;

	}

	//-i /path/to/sound_A.ogg
	if (input.cmdOptionExists("-i")) {
		const std::string& path_to_sound_A = input.getCmdOption("-i");
		std::cout << "sound_A: " << path_to_sound_A << std::endl;
		voip.setInputFile(path_to_sound_A.c_str());
		//voip.setInputFile("sound_B.ogg");
	}

	//-o /path/to/sound_output_B.ogg
	if (input.cmdOptionExists("-o")) {
		const std::string& path_to_sound_output_B = input.getCmdOption("-o");
		std::cout << "sound_output_B: " << path_to_sound_output_B << std::endl;
		voip.setOutputFile(path_to_sound_output_B.c_str());
		//voip.setOutputFile("sound_output_A");
	}

	//-c config.json
	if (input.cmdOptionExists("-c")) {
		const std::string& config = input.getCmdOption("-c");
		std::cout << "config: " << config << std::endl;

		const char* logfile = nullptr;
		if (strcmp(caller, "caller") == 0)
		{
			logfile = "caller_log.txt";
		}

		if (strcmp(caller, "callee") == 0)
		{
			logfile = "callee_log.txt";
		}
		voip.setConfigFile(config.c_str(), logfile);
		//voip.setConfigFile(config.c_str(), L"log2.txt");
		//voip.initConfig();
	}


	const IPv6Address v6addr("::0");
	vector<Endpoint> endpoints;
	int64_t id = stoll("2398853640835");
	uint16_t port = stoll(port_str.c_str());

	std::vector<char> caller2 = HexToBytes(tag_caller_hex);
	char* cal = reinterpret_cast<char*> (&caller2[0]);

	const Endpoint endpoint = Endpoint(id, port, reflector_ip, v6addr, 
		Endpoint::Type::UDP_RELAY, (unsigned char*)cal);
	endpoints.push_back(endpoint);
	

	voip.SetRemoteEndpoints(endpoints, false, tgvoip::VoIPController::GetConnectionMaxLayer());

	voip.Start();
	voip.Connect();
	
	//Sleep(10 * 1000);
	std::this_thread::sleep_for(std::chrono::milliseconds(30 * 1000));
	//Sleep(60000);

	std::cout << voip.GetDebugLog() << endl;
	voip.deinitVoIPController();
	return 0;
}
